
/**
 * Module dependencies
 */

var lunr = require('lunr')
  , file = require('./file')
  , log = require('./_log')
  , EventEmitter = require('events').EventEmitter
  , util = require('util')
  , fs = require('fs')
  , defer = require('q').defer;

/**
 * Add terminal colors
 */

require('terminal-colors');

/**
 * We need to search in both `gt search` and in the interface. This object
 * provides indexing and query functions.
 *
 * @constructor Search
 */

var Search = function() {
  this.localizations = null;
  this.index = null;
  this._createIndex();
};

/**
 * Create search index
 *
 * @return {void}
 * @api private
 */

Search.prototype._createIndex = function() {
  this.index = lunr(function() {
    this.ref('id');
    this.field('key', { boost: 10});
    this.field('value', { boost: 10});
  });
};

/**
 * We need to read all translations and append translations to the index
 *
 * @return {Promise}
 * @api public
 */

Search.prototype.readLocalizations = function() {
  var _this = this, deferred = defer();

  file.readLocalizations()
    .then(function(localizations) {
      _this.localizations = localizations;
      var language = commands.lang || project.defaultLanguage;

      if(typeof _this.localizations[language] === 'undefined') {
        throw new TypeError('Default language is undefined');
      }

      _this.localizations = _this.localizations[language];

      for(var key in _this.localizations) {
        var localization = _this.localizations[key];
        _this.index.add({
          id: localization.key,
          value: localization.value,
          // the  `BASE__EXIT_BUTTON` should be toknized as `base exit button`
          key: localization.key.replace(/_+/g,  ' ').toLowerCase()
        });
      }

      // Save docs
      _this.docs = _this.localizations;

      deferred.resolve();
    })
    .fail(function(error) {
      deferred.reject(error);
    });

  return deferred.promise;
};

/**
 * Query translations
 *
 * @param {String} q
 * @return {void}
 * @api public
 */

Search.prototype.queryOutput = function(q) {
  var _this = this;

  var res = this.index.search(q).slice(0, program.DEFAULT_LOG_LENGTH);

  if(res.length === 0) {
    log.log('No result found');
  }

  var n = 1;
  var cache = [];
  for(var i in res) {
    log.log((n === 10 ? '%' + n : ' %' + n).yellow
      + ' ' + this.docs[res[i].ref].key + ' | '
      + this.docs[res[i].ref].value.green);
    cache.push(res[i]);
    n++;
  }

  // Store as cache
  fs.writeFile(
    project.cache.search,
    JSON.stringify(cache, null, 2));
};

/**
 * Query translations
 *
 * @param {String} q
 * @return {void}
 * @api public
 */

Search.prototype.query = function(q) {
  var _this = this;

  return this.index.search(q).slice(0, program.DEFAULT_LOG_LENGTH).map(function(result) {
    return {
      id: _this.localizations[result.ref].id,
      key: result.ref,
      value: _this.localizations[result.ref].value
    };
  });
};

/**
 * Exports instance
 */

module.exports = Search;

/**
 * Exports constructor
 */

module.exports.Search = Search;
